(in-package "GWB")

(eval-when (compile load eval)
  (defparameter *strategy-debug* nil
    "If the GWB was compiled with this true, sanity checks are made while playing"))

(defmacro when-debug (&body body)
  "Expands to the body when *STRATEGY-DEBUG* is true, to nil otherwise"
  (if (not *strategy-debug*)
      'nil
    `(progn ,@body)))

(eval-when (compile load eval)
  (defun declare-fast ()
    '(declare 
      (optimize (speed 3) (safety 1) (debug 0) (compilation-speed 0)))))

(eval-when (compile load eval)
  (defconstant *strategy-types*
    '(bottom-strategy
      ground-strategy ground->ground-strategy ground3-strategy
      I-strategy K-strategy S-strategy Y-strategy
      constant-strategy un-constant-strategy composed-strategy
      conditional-strategy call/cc-strategy
      new-strategy read-strategy setter-strategy
      skip-strategy seq-strategy
      pair-strategy proj-strategy
      curried-strategy
      inj-strategy case-strategy)
    "The list of subtypes of STRATEGY"))

;;; The Eastern-European Man's Object System

(eval-when (compile load eval)
  (defun strategy-method-name (type name)
    (intern
     (concatenate 'string
                  (symbol-name name)
                  "-"
                  (symbol-name type))
     (symbol-package name))))

(defmacro def-strategy-class (type &body newslots)
  "Define a new subtype of STRATEGY"
  `(defstruct (,type (:include strategy))
    ,@newslots))

(defmacro def-strategy-generic (name (strat &rest arglist) &body decls)
  "Define a new generic function on subtypes of STRATEGY"
  (declare (symbol name))
  `(defun ,name (,strat ,@arglist)
    (declare (type strategy ,strat))
    ,@decls
    (etypecase ,strat
      ,@(mapcar
         #'(lambda (x) 
             `(,x
               (,(strategy-method-name x name) ,strat ,@arglist)))
         *strategy-types*))))

(defmacro def-strategy-method (name ((strat type) &rest arglist) &body body)
  "Define a new method on a subtype of STRATEGY"
  `(defun ,(strategy-method-name type name) ,(cons strat arglist)
    (declare (type ,type ,strat) (ignorable ,strat))
    ,@body))

(deftype proj () 'bit)
(deftype comp () 'simple-bit-vector)
(deftype token () '(or symbol simple-string number compound-token))
(deftype just () '(or null move))
(defstruct (move (:constructor make-move (comp token just)))
  (comp '#* :type comp :read-only t)
  (token nil :type token :read-only t)
  (just nil :type just :read-only t))

(deftype hist-int () '(integer 0 10000000))
(deftype type-int () '(integer 0 10000))

#+CMU (declaim (ext:freeze-type move))

(defstruct compound-token)

(deftype ground-value () '(or null integer (member true false)))

(defstruct (set-token
             (:include compound-token)
             (:constructor make-set-token (value)))
  (value nil :type ground-value))

(defstruct (value-token
             (:include compound-token)
             (:constructor make-value-token (value)))
  (value nil :type ground-value))

(defun token-equal-p (t1 t2)
  (declare (type token t1 t2))
  (cond
    ((not (or (compound-token-p t1) (compound-token-p t2)))
     (equal t1 t2))
    ((set-token-p t1) (and (set-token-p t2)
                           (equal (set-token-value t1) (set-token-value t2))))
    ((value-token-p t1) (and (value-token-p t2)
                             (equal (value-token-value t1) (value-token-value t2))))))

(defstruct strategy
  (history '() :type hist)
  (trace-stream nil :type (or null stream))
  (name nil))

(deftype hist () 'list)

(defun comp-name (c &optional (skip 0))
  "Returns the printable name of a component, skipping the first SKIP projs"
  (declare (type comp c) (type type-int skip))
  (cond
    ((< (length c) skip) (warn "No such component!") "nocomp")
    ((= (length c) skip) "")
    (t
     (do* ((len (- (length c) skip))
           (s (make-string len))
           (n 0 (+ n 1)))
          ((>= n len) s)
       (declare (type comp c) (type type-int n))
       (setf (aref s n)
             (if (zerop (aref c (+ n skip))) #\0 #\1))))))

(defun token-name (token)
  "Returns the printable name of a token"
  (declare (type token token))
  (etypecase token
    (simple-string token)
    (number (format nil "~A" token))
    (symbol (format nil "~(~A~)" token))
    (set-token (format nil "set ~A" (set-token-value token)))
    (value-token (format nil "value ~A" (value-token-value token)))))

(defun move-name (m &optional comp)
  "Returns the printable name of a move"
  (declare (type move m) (type (or list comp) comp))
  (let ((cn (comp-name (move-comp m) (length comp))))
    (if (zerop (length cn))
        (token-name (move-token m))
      (concatenate 'string
                   (token-name (move-token m))
                   cn))))
                   
