(in-package "GWB")

(defvar *move-counter* 0
  "Global move counter")
(declaim (fixnum *move-counter*))

(defmacro check-move (m s &optional errstring)
  "Does sanity checking for moves.
Compiles to nothing if *STRATEGY-DEBUG* is not set"
  (let ((m-name (gensym "m")) (s-name (gensym "s")))
    `(let ((,m-name ,m) (,s-name ,s))
      (declare (type move ,m-name) (type strategy ,s-name))
      ,@(if *strategy-debug*
          `(
            (when (not (questionp ,m-name))
              (when (or (null (move-just ,m-name))
                        (not (questionp (move-just ,m-name))))
               (cerror "Ignore this error" ,errstring)))
           (when (not (null (move-just ,m-name)))
             (when (not (find-move (move-just m-name)
                                   (strategy-history s-name)))
       (cerror "Ignore this error" errstring)))))
      ,s-name ,m-name)))

(declaim (inline do-trace))
(defun do-trace (s player move)
  "Trace a strategy"
  (declare (type strategy s))
  (declare (type move move))
  (when (strategy-trace-stream s)
    (do-trace* s player move)))

(defun do-trace* (s player move)
  "Helper function for DO-TRACE"
  (declare (type strategy s))
  (declare (type move move))
  (let ((str (strategy-trace-stream s)))
    (format str
            "~&~A ~A ~A ~A ~A ~A~%"
            (or (strategy-name s) "-")
            player
            (if (questionp move) 'q 'a)
            (if (equal '#* (move-comp move))
                "-"
              (comp-name (move-comp move)))
            (move-token move)
            (and (move-just move)
                 (move-position (move-just move) (strategy-history s))))))


(defun translate-justification (just proj1 hist1 proj2 hist2)
  "Helper for TRANSLATE-MOVE"
  ;; Proj1 or Proj2 might be nil
  (declare (type just just) (type hist hist1 hist2) (type comp proj1 proj2))
  (cond
    ((null just) nil)
    ((null hist1) (error 'gwb-translation-error))
    ((not (wherep* (car hist1) proj1))
     (translate-justification just proj1 (cdr hist1) proj2 hist2))
    ((not (wherep* (car hist2) proj2))
     (translate-justification just proj1 hist1 proj2 (cdr hist2)))
    (t
     (when-debug
      (when (not (and (eql (move-token (car hist1)) 
                           (move-token (car hist2)))
                      (equal (subseq (move-comp (car hist1)) 
                                     (length proj1))
                             (subseq (move-comp (car hist2)) 
                                     (length proj2)))))
        (error 'gwb-translation-error)))
     (if (eql just (car hist1))
         (car hist2)
       (translate-justification just
                                proj1 (cdr hist1)
                                proj2 (cdr hist2))))))

(defun translate-move (move p1 s1 p2 s2)
  "Translate a move between two strategies playing in parallel.
The components P1 of S1 and P2 of S2 have been playing in parallel;
MOVE has been played in P1 of S1: return the move to be played in P2 of S2"
  (declare (type move move) (type strategy s1 s2) (type comp p1 p2))
  (check-move move s1 "Translate-move: wrong justification on input")
  (check-move
   (changecomp move p1 p2
               (translate-justification 
                (if (null (move-just move))
                    nil
                  (justified-by (move-just move) p1))
                p1
                (if (eql move (car (strategy-history s1)))
                    (cdr (strategy-history s1))
                  (strategy-history s1))
                p2
                (strategy-history s2)))
   s2
   "Translate-move: wrong justification on output"))

(defun do-copy-move (move p1 p2 s)
  "Helper for playing copycat"
  (declare (type comp p1 p2))
  (let* ((our (move-just move))
         (his (and our (cadr (find-move our (strategy-history s))))))
    (changecomp move p1 p2
                (if (and our
                         (wherep* our p1))
                    his
                  (if (and (move-just move) (move-just (move-just move)))
                      (move-just (move-just move))
                    move)))))

(def-strategy-generic next-move (s move))

(defun play (s m)
  "Play move M against strategy S"
  (declare (type strategy s) (type move m))
  (push
   (check-move m s)
   (strategy-history s))
  (do-trace s 'o m)
  (let ((m* (next-move s m)))
    (declare (type move m*))
    (push (check-move m* s)
               (strategy-history s))
    (do-trace s 'p m*)
    (unless (typep s '(or composed-strategy
                       constant-strategy
                       un-constant-strategy))
      (incf *move-counter*))
    m*))

;;; the bottom strategy
(def-strategy-class bottom-strategy)

(def-strategy-method next-move ((s bottom-strategy) m)
  (error 'gwb-play-move-error :strategy s :move m))

;;; the top strategy
(def-strategy-class top-strategy)

(def-strategy-method next-move ((s top-strategy) m)
  (error 'gwb-play-move-error :strategy s :move m))

;;; a non-bottom constant of ground type
(def-strategy-class ground-strategy
  (value nil))

(def-strategy-method next-move ((s ground-strategy) m)
  (make-move #* (ground-strategy-value s) m))

;;; a strict function of one argument
(def-strategy-class ground->ground-strategy
  (value 
   #'(lambda (x) (declare (ignore x)) (error "uninitialized strategy"))))

(def-strategy-method next-move ((s ground->ground-strategy) m)
  (cond
   ((wherep m 1)
    (make-move #*0 *initial-question*  m))
   ((wherep m 0)
    (make-move #*1 
              (funcall (ground->ground-strategy-value s) 
                       (move-token m))
              (move-just (move-just m))))
   (t (error 'gwb-play-move-error :strategy s :move m))))

;;; a strict function of two arguments; asks the first argument first
(def-strategy-class ground3-strategy
  (value 
   #'(lambda (x y) (declare (ignore x y)) (error "uninitialized strategy"))))

(def-strategy-method next-move ((s ground3-strategy) m)
  (cond
    ((wherep m 1 1)
     (make-move #*0 *initial-question* m))
    ((wherep m 0)
     (make-move #*10
               *initial-question*
               (move-just (move-just m))))
    ((wherep m 1 0)
     (let ((m0 (cadr (find-move (move-just m) (strategy-history s)))))
       (when (not (wherep m0 0))
         (error 'gwb-play-move-error :strategy s :move m))
       (let ((result (funcall (ground3-strategy-value s)
                              (move-token m0)
                              (move-token m))))
         (make-move
          #*11
          result
          (move-just (move-just m))))))
    (t (error 'gwb-play-move-error :move m :strategy s))))

;;; an identity strategy
(def-strategy-class I-strategy)

(def-strategy-method next-move ((s I-strategy) m)
  (do-copy-move m (where-comp m) (where-comp m t) s))

;;; the strategy for the K combinator
(def-strategy-class K-strategy)

(def-strategy-method next-move ((s K-strategy) m)
  (when (wherep m 1 0)
    (error 'gwb-play-move-error :strategy s :move m))
  (multiple-value-bind 
        (from to)
      (if (wherep m 1 1) (values #*11 #*0) (values #*0 #*11))
    (do-copy-move m from to s)))

;;; the strategy for the S combinator \xyz.(xz)(yz)
(def-strategy-class s-strategy)

(def-strategy-method next-move ((s S-strategy) m)
  (cond
    ((wherep m 1 1 1)
     (do-copy-move m #*111 #*011 s))
    ((wherep m 0 1 1)
     (do-copy-move m #*011 #*111 s))
    ((wherep m 1 0 1)
     (do-copy-move m #*101 #*010 s))
    ((wherep m 0 1 0)
     (do-copy-move m #*010 #*101 s))
    ((wherep m 0 0)
     (do-copy-move m #*00 #*110 s))
    ((wherep m 1 0 0)
     (do-copy-move m #*100 #*110 s))
    ((wherep m 1 1 0)
     (let ((n* (move-just m)))
       (when (null n*)
         (error 'gwb-play-move-error :strategy s :move m))
       (let ((n (cadr (find-move n* (strategy-history s)))))
         (cond
           ((null n)
            (error 'gwb-play-move-error :strategy s :move m))
           ((wherep n 0 0) 
            (do-copy-move m #*110 #*00 s))
           ((wherep n 1 0 0)
            (do-copy-move m #*110 #*100 s))
           (t (error 'gwb-play-move-error :strategy s :move m))))))
    (t (error 'gwb-play-move-error :strategy s :move m))))

;;; the Y combinator
(def-strategy-class Y-strategy)

;;; a move in 1 is copied into 01
;;; a move in 01 is copied into 1 if hereditarily justified in 1
;;;                             00                             00
(def-strategy-method next-move ((s Y-strategy) m)
  (cond
    ((wherep m 1)
     (do-copy-move m #*1 #*01 s))
    ((wherep m 0 0)
     (do-copy-move m #*00 #*01 s))
    ((wherep m 0 1)
     (let ((n (move-just m)))
       (when (null n)
         (error 'gwb-play-error :strategy s))
       (let ((n* (cadr (find-move n (strategy-history s)))))
         (when (null n*)
           (error 'gwb-play-move-error :strategy s :move m))
         (cond
           ((wherep n* 1)
            (do-copy-move m #*01 #*1 s))
           ((wherep n* 0 0)
            (do-copy-move m #*01 #*00 s))
           (t (error 'gwb-play-move-error :strategy s :move m))))))
    (t (error 'gwb-play-move-error :strategy s :move m))))

;;; call/cc
(def-strategy-class call/cc-strategy)

(defun initial-move (s)
  (let ((m (car s)))
    (if (null (move-just m))
        m
      (initial-move (find-move (move-just m) s)))))

(def-strategy-method next-move ((s call/cc-strategy) m)
  (cond
    ((wherep m 1)
     (let ((m* (move-just m)))
       (if (null m*)
           (do-copy-move m #*1 #*01 s)
         (let ((m** (cdr (find-move m* (strategy-history s)))))
           (cond
             ((wherep m** 0 1)
              (do-copy-move m #*1 #*01 s))
             ((wherep m** 0 0 0)
              (do-copy-move m #*1 #*000 s))
             (t (error 'gwb-play-move-error :strategy s :move m)))))))
    ((wherep m 0 1)
     (do-copy-move m #*01 #*11 s))
    ((wherep m 0 0 1)
     (make-move #*000 *initial-question* m))
    ((wherep m 0 0 0)
     (if (wherep (move-just (move-just m)) 0 0 1)
         (changecomp m #*000 #*1
                     (initial-move (strategy-history s)))
       (do-copy-move m #*000 #*1 s)))))

;;; a (non-strict) constant strategy; possibly higher-order
(def-strategy-class constant-strategy
    (value nil))

(def-strategy-method next-move ((s constant-strategy) m)
  (let ((real-s (constant-strategy-value s)))
    (let ((n (play real-s 
                   (translate-move m (where-comp m) s #* real-s))))
      (translate-move n #* real-s #*1 s))))

;;; the value of a non-strict constant strategy
(def-strategy-class un-constant-strategy
    (value nil))

(def-strategy-method next-move ((s un-constant-strategy) m)
  (let ((real-s (un-constant-strategy-value s)))
    (let ((n (play real-s (translate-move m #* s #*1 real-s))))
      (when (not (wherep n 1))
        (error 'gwb-play-move-error :strategy s :move m))
      (translate-move n #*1 real-s #* s))))


;;; the composition of two strategies
(def-strategy-class composed-strategy
  (s1 nil)
  (s2 nil))

;;; this implements a sort of parallel composition
(def-strategy-method next-move ((s composed-strategy) m)
  (labels
      ((intercourse (s1 s2 m)
         (declare (type strategy s1 s2) (type move m))
         (multiple-value-bind 
               (s-last s-next lastcomp nextcomp)
             (if (wherep m 0) 
                 (values s2 s1 #*0 #*1)
               (values s1 s2 #*1 #*0))
           (let ((m* (translate-move m 
                                     lastcomp s-last 
                                     nextcomp s-next)))
             (let ((n (play s-next m*)))
               (if (eql (comp-car lastcomp) (where n))
                   n
                 (intercourse s1 s2 n)))))))
    ;; main body
    (let ((s1 (composed-strategy-s1 s)) (s2 (composed-strategy-s2 s)))
      (multiple-value-bind 
            (new-s comp)
          (if (wherep m 1) 
              (values s2 #*1)
            (values s1 #*0))
        (let ((n* (play new-s
                        (translate-move m 
                                        comp s 
                                        comp new-s))))
          (let ((n (if (eql (comp-car comp) (where n*))
                       n*
                     (intercourse s1 s2 n*))))
            ;; now n is in |a or |c
            (multiple-value-bind 
                  (last-s lastcomp)
                (if (wherep n 0) 
                    (values s1 #*0)
                  (values s2 #*1))
              (translate-move n
                              lastcomp last-s
                              lastcomp s))))))))

;;; the conditional, of type o->a->a->a
(def-strategy-class conditional-strategy)

(def-strategy-method next-move ((s conditional-strategy) m)
  (cond
    ((wherep m 0)                      ; a boolean value
     (let ((n (move-just (move-just m)))
           (val (move-token m)))
       (ecase val
         (true
          (make-move #*10 *initial-question* n))
         (false
          (make-move #*110 *initial-question* n)))))
    ((wherep m 1 1 1)                  ; result component
     (let ((n (move-just m)))
       (cond
         ((null n)
          (make-move #*0 *initial-question* m))
         (t
          (let ((n-hist (find-move n (strategy-history s))))
            (when (null n-hist)
              (error 'gwb-play-move-error :strategy s :move m))
            (cond
              ((wherep (cadr n-hist) 1 0)
               (do-copy-move m #*111 #*10 s))
              ((wherep (cadr n-hist) 1 1 0)
               (do-copy-move m #*111 #*110 s))
              (t (error 'gwb-play-move-error :strategy s :move m))))))))
    (t                                ; we're in a branch of the alternative
     (let ((from (if (wherep m 1 0) #*10 #*110)))
       (if (initial-move-p (move-just (move-just m)))
           ;; can't use do-copy-move
           (changecomp m from #*111
                       (move-just (move-just m)))
         (do-copy-move m 
           (if (wherep m 1 0) #*10 #*110)
           #*111
           s))))))

(def-strategy-class new-strategy
    (value nil))

(def-strategy-method next-move ((s new-strategy) m)
  (unless (equal (move-comp m) #*)
    (error 'gwb-play-move-error :strategy s :move m))
  (let ((tok (move-token m)))
    (cond
      ((eql tok 'read)
       (when (null (new-strategy-value s))
         (error "Uninitialized reference"))
       (make-move #* (new-strategy-value s) m))
      ((set-token-p tok)
       (setf (new-strategy-value s)
             (make-value-token (set-token-value tok)))
       (make-move #* 'ok m))
      (t (error 'gwb-play-move-error :move m :strategy s)))))

(def-strategy-class read-strategy)

(def-strategy-method next-move ((s read-strategy) m)
  (cond
    ((wherep m 1) (make-move #*0 'read m))
    ((wherep m 0)
     (make-move #*1 (value-token-value (move-token m))
                (move-just (move-just m))))
    (t (error 'gwb-play-move-error :move m :strategy s))))

(def-strategy-class setter-strategy)

(def-strategy-method next-move ((s setter-strategy) m)
  (cond
    ((wherep m 1 1) (make-move #*10 *initial-question* m))
    ((wherep m 1 0)
     (let ((value (move-token m)))
       (make-move #*0 (make-set-token value)
                  (move-just (move-just m)))))
    ((wherep m 0)
     (make-move #*11 '* (move-just (move-just m))))
    (t (error 'gwb-play-move-error :move m :strategy s))))

(def-strategy-class skip-strategy)

(def-strategy-method next-move ((s skip-strategy) m)
  (make-move #* '* m))
  
(def-strategy-class seq-strategy)

(def-strategy-method next-move ((s seq-strategy) m)
  (cond
    ((wherep m 1 1)
     (let ((n (move-just m)))
       (cond
         ((null n) (make-move #*0 *initial-question* m))
         (t
          (do-copy-move m #*11 #*10 s)))))
    ((wherep m 0)
     (let ((n (move-just (move-just m))))
       (changecomp n #*11 #*10 n)))
    ((wherep m 1 0)
     (if (initial-move-p (move-just (move-just m)))
         (changecomp m #*10 #*11 (move-just (move-just m)))
       (do-copy-move m #*10 #*11 s)))
    (t (error 'gwb-play-move-error :move m :strategy s))))
  
(def-strategy-class pair-strategy)

(def-strategy-method next-move ((s pair-strategy) m)
  (cond
    ((wherep m 1 1)
     (cond            
       ((wherep m 1 1 0)
        (do-copy-move m #*110 #*0 s))
       ((wherep m 1 1 1)
        (do-copy-move m #*111 #*10 s))
       (t
        (error 'gwb-play-move-error :strategy s :move m))))
    ((wherep m 0)
     (do-copy-move m #*0 #*110 s))
    ((wherep m 1 0)
     (do-copy-move m #*10 #*111 s))
    (t (error 'gwb-play-move-error :move m :strategy s))))
     
(def-strategy-class proj-strategy
    (component 0))

(defun make-p0-strategy (&rest args)
  (apply #'make-proj-strategy :component 0 args))

(defun make-p1-strategy (&rest args)
  (apply #'make-proj-strategy :component 1 args))

(def-strategy-method next-move ((s proj-strategy) m)
  (let ((component (proj-strategy-component s)))
    (cond
      ((wherep m 1)
       (let ((new-component
              (coerce
               (vector 0 (proj-strategy-component s))
               'comp)))
         (do-copy-move m #*1 new-component s)))
      ((wherep m 0)
       (let ((comp (coerce (vector 0 component) 'comp)))
         (if (wherep* m comp)
             (do-copy-move m comp #*1 s)
           (error 'gwb-play-move-error :strategy s :move m))))
      (t (error 'gwb-play-move-error :move m :strategy s)))))

(def-strategy-class curried-strategy
    (value (make-bottom-strategy)))

(def-strategy-method next-move ((s curried-strategy) m)
  (let ((real-s (curried-strategy-value s)))
    (multiple-value-bind (from to)
        (cond
          ((wherep m 0) (values #*0 #*00))
          ((wherep m 1 0) (values #*10 #*01))
          ((wherep m 1 1) (values #*11 #*1))
          (t (error 'gwb-play-move-error :strategy s :move m)))
      (let ((n (play real-s (translate-move m from s to real-s))))
        (multiple-value-bind (from to)
            (cond
              ((wherep n 1) (values #*1 #*11))
              ((wherep n 0 1) (values #*01 #*10))
              ((wherep n 0 0) (values #*00 #*0))
              (t (error 'gwb-play-move-error :strategy s :move m)))
          (translate-move n from real-s to s))))))
     
(def-strategy-class inj-strategy
    (component 0))

(def-strategy-method next-move ((s inj-strategy) m)
  (declare (ignore m))
  (error "Not implemented"))

(def-strategy-class case-strategy)

(def-strategy-method next-move ((s case-strategy) m)
  (declare (ignore m))
  (error "Not implemented"))

;;; various functions for building strategies


(defun fresh-strategy (s)
  "Make a copy of a strategy with a virgin history"
  (declare (type strategy s))
  (let ((ss (copy-strategy s)))
    (setf (strategy-history ss) '())
    (setf (strategy-name ss) nil)
    (setf (strategy-trace-stream ss) nil)
    (typecase ss
      (constant-strategy
       (setf (constant-strategy-value ss) 
             (fresh-strategy (constant-strategy-value ss))))
      (un-constant-strategy
       (setf (un-constant-strategy-value ss)
             (fresh-strategy (un-constant-strategy-value ss))))
      (composed-strategy
       (setf (composed-strategy-s1 ss)
             (fresh-strategy (composed-strategy-s1 ss)))
       (setf (composed-strategy-s2 ss)
             (fresh-strategy (composed-strategy-s2 ss))))
      (new-strategy
       (setf (new-strategy-value ss) (make-value-token 0))))
    ss))
  
(defun un-constant-strategy (s)
  (declare (type strategy s))
  (make-un-constant-strategy
   :value s))

(defun constant-strategy (s)
  (declare (type strategy s))
  (make-constant-strategy :value s))

(defun composed-strategy (s1 s2)
  (declare (type strategy s1 s2))
  (make-composed-strategy
   :s1 s1 :s2 s2))

(defun apply-strategy (f x)
  (declare (type strategy f x))
  (un-constant-strategy 
   (composed-strategy (constant-strategy x) f)))

(defparameter *strategy-trace* *terminal-io*
  "The default stream for tracing strategies")

(defun trace-strategy (s name &optional (stream *strategy-trace*))
  (declare (special *strategy-trace*))
  (declare (type strategy s))
  (setf (strategy-trace-stream s) stream)
  (setf (strategy-name s) name)
  s)

(defmacro with-trace-to-file (filename &body body)
  `(with-open-file (*strategy-trace* 
                    ,filename 
                    :direction :output
                    :if-exists :supersede)
    ,@body
    (format *strategy-trace* "~&~S ~S ~S ~S ~S~&" nil nil nil nil nil)))

(defun run-strategy (s)
  (declare (type strategy s))
  (setf *move-counter* 0)
  (play s (make-move #* *initial-question* nil))
  *move-counter*)

(defun strategy-returned-value (s)
  "Return the last value returned by a strategy"
  (declare (type strategy s))
  (move-token (car (strategy-history s))))
